"""Addtional tools for `graphene_django`.  """
import sys
from typing import Type, Union

import django
import graphene
import graphene_django
from graphene.types.mountedtype import MountedType
from graphene.types.unmountedtype import UnmountedType
from graphene_django.registry import get_global_registry
from graphql import GraphQLError


def get_modelnode(model: Type[django.db.models.Model], is_autocreate=True)\
        -> Type[graphene_django.DjangoObjectType]:
    """Get graphene node class from model class.

    Args:
        model (Type[django.db.models.Model])
        is_autocreate (Boolean): Create modelnode type
            use default config if not created.

    Returns:
        Type[graphene_django.DjangoObjectType]
    """

    assert issubclass(model, django.db.models.Model), type(model)
    registry = get_global_registry()
    ret = registry.get_type_for_model(model)
    if ret:
        return ret
    if is_autocreate:
        return create_modelnode(model)
    raise RuntimeError(
        f'Defined node type first, or set `is_autocreate` to True: {model}')


def create_modelnode(model: Type[django.db.models.Model], bases=(), **meta_options) \
        -> Type[graphene_django.DjangoObjectType]:
    """Create node class for model class.

    Raises:
        AssertionError: node type already created.

    Returns:
        Type: Node type for model
    """

    assert issubclass(model, django.db.models.Model), type(model)

    from . import texttools
    meta_options.setdefault('model', model)
    meta_options.setdefault('interfaces', (graphene.Node,))
    meta_options.setdefault(
        'description', f'Auto created node type from model: {model.__name__}')
    if graphene_django.utils.DJANGO_FILTER_INSTALLED:
        meta_options.setdefault('filter_fields', '__all__')
    bases += (graphene_django.DjangoObjectType,)
    clsname = texttools.camel_case(f'{model.__name__}_node')
    return type(clsname, bases, dict(Meta=meta_options))


def handle_resolve_error():
    """Detail message for `graphql.error.located_error.GraphQLLocatedError`.  """

    type_, value, _ = sys.exc_info()
    if not isinstance(value, GraphQLError):
        import traceback
        traceback.print_exc()
        raise GraphQLError(f'{type_.__name__}:{value}')
    else:
        raise value


def get_node_id(instance: django.db.models.Model) -> str:
    """Get instance global node id.

    Args:
        instance (django.db.models.Model): Model instance

    Returns:
        str: Global id.
    """

    assert isinstance(instance, django.db.models.Model), type(instance)
    modelnode = get_modelnode(instance.__class__, is_autocreate=False)
    return graphene.Node.to_global_id(modelnode.__name__, instance.pk)


def get_unmounted_type(obj: Union[MountedType, UnmountedType]) -> UnmountedType:
    """Get unmounted type of given object object.  

    Args:
        obj (Union[MountedType, UnmountedType]): Graphene object

    Returns:
        UnmountedType
    """

    unmounted = obj.type if isinstance(MountedType) else obj
    if isinstance(unmounted, graphene.NonNull):
        unmounted = unmounted.of_type
    return unmounted




"""Mutation base for django model  """


import re
from collections import OrderedDict
from collections.abc import Callable
from typing import Dict, Iterable, Tuple

import django
import graphene
from django.db import models, transaction
from graphene.types.unmountedtype import UnmountedType
from graphene_django.registry import Registry, get_global_registry
from graphql import GraphQLError

from . import core
from ..converter import convert_db_field_to_argument
from ..types import ModelField, ModelListField
from .node import NodeMutation, NodeUpdateMutation


def get_all_fields(model):
    """Get all fields from model."""

    # pylint: disable=protected-access
    meta: django.db.models.options.Options = model._meta
    return meta.get_fields()


def construct_argument_fields(
        model: models.Model,
        registry: Registry,
        include: Iterable[str],
        require: Iterable[str])-> OrderedDict:
    """Construct fields for argument.

    Args:
        model (models.Model): Django db model.
        registry (Registry): Graphene django registry.
        include (Iterable[str]): Include field names.
        require (Iterable[str]): Required(Non-null) fields names.

    Returns:
        OrderedDict: Fields
    """

    fields = [i for i in get_all_fields(model) if i.name in include]

    ret = OrderedDict()
    for i in fields:
        converted = convert_db_field_to_argument(i, registry)
        if not converted:
            continue
        converted.kwargs['required'] = i.name in require
        ret[i.name] = converted
    return ret


class ModelMutation(NodeMutation):
    """Mutate django model.  """
    # pylint: disable=abstract-method
    class Meta:
        abstract = True

    @classmethod
    def __init_subclass_with_meta__(cls, **options):
        options.setdefault('_meta', core.ModelMutationOptions(cls))
        options.setdefault('require', ())
        options.setdefault('exclude', ())
        options.setdefault('require_mapping', True)
        options.setdefault('registry', get_global_registry())
        assert isinstance(options['registry'], Registry)
        if(not isinstance(options.get('fields'), (Callable, Iterable))):
            raise ValueError(f'Meta options for `{cls.__name__}` require'
                             ' `fields` to be either callable or iterable.')
        super().__init_subclass_with_meta__(**options)

    @classmethod
    def _construct_meta(cls, **options) -> core.ModelMutationOptions:
        ret = super()._construct_meta(**options)  # type: core.ModelMutationOptions
        ret.model = options['model']
        ret.registry = options['registry']
        return ret

    @classmethod
    def _make_arguments_fields(cls, **options) -> OrderedDict:
        options.setdefault('arguments', {})

        options['arguments'].update(
            cls._make_model_arguments_fields(**options))
        ret = super()._make_arguments_fields(**options)
        return ret

    @classmethod
    def _make_model_arguments_fields(cls, **options) -> dict:
        model = options['model']

        field_dict = OrderedDict(
            __doc__=f'Mapping data for mutation: {cls.__name__}',
            Meta=dict())
        field_dict.update(cls._get_model_input_fields(**options))
        field_objecttype = type(
            re.sub("Mapping$|$", "Mapping", cls.__name__),
            (graphene.InputObjectType,),
            field_dict)
        field = field_objecttype(required=options['require_mapping'],
                                 description=f'Mapping data for model: {model.__name__}')

        return {'mapping': field}

    @classmethod
    def _make_payload_fields(cls, **options) -> OrderedDict:
        model = options['model']  # type: django.db.models.Model
        ret = super()._make_payload_fields(**options)
        ret[options['node_fieldname']] = ModelField(model)
        return ret

    @classmethod
    def _make_context(cls, root, info: core.ResolveInfo, **kwargs):
        arguments = kwargs['input']
        return core.ModelMutationContext(
            root=root,
            info=info,
            node=None,
            options=cls._meta,
            arguments=arguments,
            mapping=None)

    @classmethod
    def _get_model_input_fields(cls, **options)-> Dict[str, UnmountedType]:

        model = options['model']
        fields = options['fields']
        require = options['require']
        exclude = options['exclude']
        if callable(fields):
            fields = fields(model)
            fields = list(i.name if hasattr(i, 'name') else i
                          for i in fields)
        fields = [i for i in fields if i not in exclude]
        ret = construct_argument_fields(model,
                                        options['registry'],
                                        fields,
                                        require,)
        ret = graphene.types.utils.yank_fields_from_attrs(ret)
        return ret

    @classmethod
    def _get_model_fields(cls, **options) -> list:
        model = options['model']  # type: django.db.models.Model
        fields = options['fields']
        return (i for i in get_all_fields(model) if i.name in fields)

    @classmethod
    def premutate(cls, context: core.ModelMutationContext):
        super().premutate(context)
        if hasattr(context, 'mapping'):
            context.mapping = context.arguments['mapping']

    @classmethod
    def mutate(cls, context: core.ModelMutationContext)\
            -> graphene.ObjectType:

        return cls(**{context.options.node_fieldname: context.instance})

    @classmethod
    def postmutate(cls,
                   context: core.ModelMutationContext,
                   payload: graphene.ObjectType) -> graphene.ObjectType:
        ret = super().postmutate(context, payload)
        try:
            context.instance.save()
        except AttributeError:
            pass
        return ret

    @classmethod
    def sorted_mapping(cls,
                       context: core.ModelMutationContext,
                       mapping: dict) -> Tuple[dict, dict]:
        """Get sorted mapping by type.

        Args:
            context (core.ModelMutationContext): Mutation context

        Raises:
            ValueError: Try use auto field in mapping.

        Returns:
            Tuple[dict, dict]: normal mapping, many to many mapping
        """

        fields = get_all_fields(context.options.model)
        normal_mapping = dict(mapping)
        m2m_mapping = {}
        for i in fields:
            if i.name not in mapping:
                continue
            if isinstance(i, django.db.models.AutoField):
                raise ValueError('Assign to auto field is not allowed')
            elif isinstance(i, django.db.models.ManyToManyField):
                m2m_mapping[i.name] = normal_mapping.pop(i.name)

        return normal_mapping, m2m_mapping


class ModelCreationMutation(ModelMutation):
    """Create model.  """

    class Meta:
        abstract = True

    @classmethod
    def mutate(cls, context: core.ModelMutationContext)\
            -> graphene.ObjectType:
        context.instance = cls.create(context, context.mapping)
        return super().mutate(context)

    @classmethod
    def create(cls,
               context: core.ModelMutationContext,
               mapping: dict) -> django.db.models.Model:
        """Create instance from mapping.  """

        normal_mapping, m2m_mapping = cls.sorted_mapping(context, mapping)
        instance = context.options.model(**normal_mapping)
        instance.save()
        for k, v in m2m_mapping.items():
            getattr(instance, k).set(v)
        return instance


class ModelUpdateMutation(NodeUpdateMutation, ModelMutation):
    """Update model.  """

    class Meta:
        abstract = True

    @classmethod
    def _construct_meta(cls, **options) -> core.ModelUpdateMutationOptions:
        options.setdefault('_meta', core.ModelUpdateMutationOptions(cls))
        ret = super()._construct_meta(**options) \
            # type: core.ModelUpdateMutationOptions
        return ret

    @classmethod
    def premutate(cls, context: core.ModelMutationContext):

        super().premutate(context)
        if not (isinstance(context.node, context.options.model)
                or (isinstance(context.node, list)
                    and all(isinstance(i, context.options.model) for i in context.node))):
            raise GraphQLError(
                f'Got a {type(context.node)} node, expected for {context.options.model}')
        context.instance = context.node

    @classmethod
    def mutate(cls, context: core.ModelMutationContext):
        normal_mapping, m2m_mapping = cls.sorted_mapping(
            context, context.mapping)
        for k, v in normal_mapping.items():
            setattr(context.instance, k, v)
        for k, v in m2m_mapping.items():
            getattr(context.instance, k).set(v)
        return super().mutate(context)


class ModelBulkMutation(ModelMutation):
    class Meta:
        abstract = True

    @classmethod
    def __init_subclass_with_meta__(cls, **options):
        options.setdefault('node_fieldname', 'node_set')
        super().__init_subclass_with_meta__(**options)

    @classmethod
    def _make_payload_fields(cls, **options) -> OrderedDict:
        model = options['model']  # type: django.db.models.Model
        ret = super()._make_payload_fields(**options)
        node_fieldname = options['node_fieldname']
        ret[node_fieldname] = ModelListField(model).Field()
        return ret


class ModelBulkCreationMutation(ModelCreationMutation, ModelBulkMutation):
    """Create model in bulk.  """

    class Meta:
        abstract = True

    @classmethod
    def _make_context(cls, root, info: core.ResolveInfo, **kwargs):
        arguments = kwargs['input']
        return core.ModelBulkCreationMutationContext(
            root=root,
            info=info,
            options=cls._meta,
            arguments=arguments,
            data=None)

    @classmethod
    def premutate(cls, context: core.ModelBulkCreationMutationContext):
        super().premutate(context)
        context.data = context.arguments['data']

    @classmethod
    def _make_model_arguments_fields(cls, **options) -> dict:
        ret = super()._make_model_arguments_fields(**options)
        ret['data'] = graphene.List(ret.pop('mapping').__class__)
        return ret

    @classmethod
    def mutate(cls, context: core.ModelBulkCreationMutationContext)\
            -> graphene.ObjectType:
        ret = []
        for i in context.data:
            ret.append(cls.create(context, i))
        return cls(**{context.options.node_fieldname: ret})


class ModelBulkUpdateMutation(ModelUpdateMutation, ModelBulkMutation):
    """Update model in bulk.  """

    class Meta:
        abstract = True

    @classmethod
    def _make_node_arguments_fields(cls, **options) -> dict:
        id_fieldname = options['id_fieldname']
        ret = super()._make_node_arguments_fields(**options)
        ret[id_fieldname] = graphene.List(ret[id_fieldname].__class__).Field()
        return ret

    @classmethod
    def premutate(cls, context: core.ModelBulkUpdateMutationContext):
        super().premutate(context)
        assert isinstance(context.node, list) and all(isinstance(
            i, context.options.model) for i in context.node)
        context.query_set = context.options.model.objects.filter(
            pk__in=(i.pk for i in context.node))

    @classmethod
    @transaction.atomic
    def mutate(cls, context: core.ModelMutationContext):
        normal_mapping, m2m_mapping = cls.sorted_mapping(
            context, context.mapping)
        context.query_set.update(**normal_mapping)
        for i in context.query_set:
            for k, v in m2m_mapping.items():
                getattr(i, k).set(v)
            i.save()
        return cls(**{context.options.node_fieldname: context.query_set})



"""Mixin for predefined fields.  """

import graphene


class Message(graphene.Interface):
    """Mutation with a message.  """

    message = graphene.String(description='Human-readable execution status.')


class ClientMutationID(graphene.Interface):
    """Mutation with a client mutation id.  """

    class Arguments:
        client_mutation_id = graphene.String()

    client_mutation_id = graphene.String()




"""Additional tools for `graphene_django`.  """

from graphql.execution.base import ResolveInfo

from . import interfaces, utils
from .core import create_modelnode, get_modelnode, get_node_id
from .dataloader import DataLoaderModelFilterConnectionField
from .mutation import (ModelBulkCreationMutation,
                       ModelBulkCreationMutationContext, ModelBulkMutation,
                       ModelBulkUpdateMutation, ModelBulkUpdateMutationContext,
                       ModelCreationMutation, ModelMutation,
                       ModelMutationContext, ModelUpdateMutation, Mutation,
                       MutationContext, NodeDeleteMutation,
                       NodeDeleteMutationContext, NodeMutation,
                       NodeMutationContext, NodeUpdateMutation,
                       NodeUpdateMutationContext, get_all_fields)
from .types import (CountableConnection, Duration, ModelConnectionField,
                    ModelField, ModelFilterConnectionField, ModelListField,
                    ModelNode)



"""Graphql types for data loader.  """

from functools import partial

import graphene
from django.db import models
from graphene_django.utils import maybe_queryset
from graphql.execution.base import ResolveInfo
from graphql_relay.connection.arrayconnection import connection_from_list_slice
from promise import Promise

from ..types import ModelFilterConnectionField


class DataLoaderModelFilterConnectionField(ModelFilterConnectionField):
    """Model filter connection field with data loader support.  """

    @classmethod
    def resolve_connection(cls, connection, default_manager, args, iterable):
        def construct_connection(iterable):
            iterable = list(iterable)
            _len = len(iterable)
            ret = connection_from_list_slice(
                iterable,
                args,
                slice_start=0,
                list_length=_len,
                list_slice_length=_len,
                connection_type=connection,
                edge_type=connection.Edge,
                pageinfo_type=graphene.relay.PageInfo,
            )
            ret.iterable = iterable
            ret.length = _len
            return ret

        info = getattr(iterable, 'info', None)
        if iterable is None:
            iterable = default_manager
        iterable = maybe_queryset(iterable)

        if isinstance(iterable, models.QuerySet) and iterable is not default_manager:
            default_queryset = maybe_queryset(default_manager)
            iterable = cls.merge_querysets(default_queryset, iterable)
        if info:
            iterable = (info
                        .context
                        .get_data_loader(default_manager
                                         .model)
                        .load_many(iterable
                                   .values_list('pk', flat=True)))
        if Promise.is_thenable(iterable):
            return Promise.resolve(iterable).then(construct_connection)
        return construct_connection(iterable)

    def get_resolver(self, parent_resolver):
        def resolver(parent, info: ResolveInfo, **kwargs):
            ret = parent_resolver(parent, info, **kwargs)
            try:
                ret.info = info
            except AttributeError:
                # `ret` may be None
                pass
            return ret

        return partial(
            self.connection_resolver,
            resolver,
            self.type,
            self.get_manager(),
            self.max_limit,
            self.enforce_first_or_last,
            self.filterset_class,
            self.filtering_args
        )



"""Graphene types.  """

from datetime import timedelta
from functools import partial

import graphene
import graphene_django
import isodate
from graphene_django.filter import DjangoFilterConnectionField
from graphql.language import ast
from promise import Promise

from . import core


class ModelField(graphene.Field):
    """`Field` for model.

    Use `create_nodemodel` first if you want customize the node.
    """

    def __init__(self, model, **kwargs):
        kwargs.setdefault('description',
                          (model.__doc__
                           or f'Node for database model: {model.__name__}'))
        super().__init__(lambda: core.get_modelnode(model), **kwargs)


class ModelListField(graphene.List):
    """`List` field for model.

    Use `create_nodemodel` first if you want customize the node.
    """

    def __init__(self, model, **kwargs):
        kwargs.setdefault('description',
                          (model.__doc__
                           or f'Node list for database model: {model.__name__}'))
        super().__init__(lambda: core.get_modelnode(model), **kwargs)


class CustomConnectionResolveMixin:
    """Print traceback when encountered a error"""

    @classmethod
    def connection_resolver(cls, *args, **kwargs):
        """Resolve connection.  """
        try:
            return super().connection_resolver(*args, **kwargs)
        except:
            core.handle_resolve_error()
            raise


class ModelConnectionField(CustomConnectionResolveMixin, graphene_django.DjangoConnectionField):
    """`DjangoConnectionField` for model.

    Use `create_nodemodel` first if you want customize the node.
    """

    def __init__(self, model, **kwargs):
        kwargs.setdefault('description',
                          (model.__doc__
                           or f'Connection for database model: {model.__name__}'))
        super().__init__(lambda: core.get_modelnode(model), **kwargs)

    @classmethod
    def connection_resolver(
            cls,
            resolver,
            connection,
            default_manager,
            max_limit,
            enforce_first_or_last,
            root,
            info,
            **kwargs):
        # pylint: disable=R0913,W0221

        first = kwargs.get("first")
        last = kwargs.get("last")
        if not (first is None or first > 0):
            raise ValueError(
                f'`first` argument must be positive, got `{first}`')
        if not (last is None or last > 0):
            raise ValueError(
                f'`last` argument must be positive, got `{last}`')
        if enforce_first_or_last and not (first or last):
            raise ValueError(
                f"You must provide a `first` or `last` value "
                f"to properly paginate the `{info.field_name}` connection.")

        if not max_limit:
            pass
        elif first is None and last is None:
            kwargs['first'] = max_limit
        else:
            count = min(i for i in (first, last) if i)
            if count > max_limit:
                raise ValueError(f"Requesting {count} records "
                                 f"on the `{info.field_name}` connection "
                                 f"exceeds the limit of {max_limit} records.")

        iterable = resolver(root, info, **kwargs)
        on_resolve = partial(cls.resolve_connection,
                             connection, default_manager, kwargs)

        if Promise.is_thenable(iterable):
            return Promise.resolve(iterable).then(on_resolve)

        return on_resolve(iterable)


class ModelFilterConnectionField(ModelConnectionField):
    """`DjangoFilterConnectionField` for model.  """

    args = DjangoFilterConnectionField.args
    filterset_class = DjangoFilterConnectionField.filterset_class
    filtering_args = DjangoFilterConnectionField.filtering_args
    merge_querysets = DjangoFilterConnectionField.merge_querysets
    get_resolver = DjangoFilterConnectionField.get_resolver

    def __init__(self, model,
                 fields=None,
                 extra_filter_meta=None,
                 filterset_class=None,
                 **kwargs):
        kwargs.setdefault('description',
                          (model.__doc__
                           or f'Filterable connection for database model: {model.__name__}'))
        self._node_filterset_class = None

        self._fields = fields
        self._provided_filterset_class = filterset_class
        self._filterset_class = None
        self._extra_filter_meta = extra_filter_meta
        self._base_args = None
        super().__init__(model, **kwargs)

    @property
    def _provided_filterset_class(self):
        if self._node_filterset_class is None:
            meta = getattr(self.node_type, '_meta')
            self._node_filterset_class = getattr(meta, 'filterset_class', None)
        return self._node_filterset_class

    @_provided_filterset_class.setter
    def _provided_filterset_class(self, value):
        self._node_filterset_class = value

    @classmethod
    def connection_resolver(
            cls,
            resolver,
            connection,
            default_manager,
            max_limit,
            enforce_first_or_last,
            filterset_class,
            filtering_args,
            root,
            info,
            **kwargs):
        # pylint: disable=R0913,W0221
        filter_kwargs = {k: v for k,
                         v in kwargs.items()
                         if k in filtering_args}
        queryset = filterset_class(
            data=filter_kwargs,
            queryset=default_manager.get_queryset(),
            request=info.context,).qs

        return super().connection_resolver(
            resolver,
            connection,
            queryset,
            max_limit,
            enforce_first_or_last,
            root,
            info,
            **kwargs)


class Duration(graphene.Scalar):
    """Duration in ISO-8601 format.  """

    @staticmethod
    def serialize(duration: timedelta):
        """Serialize python object.

        Args:
            duration (timedelta): Duration

        Returns:
            str
        """

        return isodate.duration_isoformat(duration)

    @classmethod
    def parse_literal(cls, node):
        """Parse ast node.

        Args:
            node: AST node

        Returns:
            timedelta | None
        """

        if isinstance(node, ast.StringValue):
            return cls.parse_value(node.value)
        return None

    @staticmethod
    def parse_value(value):
        """Parse str to python object.

        Args:
            value (str): Value

        Returns:
            timedelta | None
        """

        try:
            return isodate.parse_duration(value)
        except ValueError:
            return None


class CountableConnection(graphene.relay.Connection):
    """Extended connection with total count.  """

    class Meta:
        abstract = True

    total = graphene.Int()

    @staticmethod
    def resolve_total(root, _info):
        return root.length


class ModelNodeOptions(graphene_django.types.DjangoObjectTypeOptions):
    """Extended DjangoObjectTypeOptions for `ModelNode`.  """

    filterset_class = None


class ModelNode(graphene_django.DjangoObjectType):
    class Meta:
        abstract = True

    @classmethod
    def __init_subclass_with_meta__(
            cls,
            filterset_class=None,
            **options):
        # pylint: disable=W0221
        _meta = options.get('_meta', ModelNodeOptions(cls))
        _meta.filterset_class = filterset_class
        options['_meta'] = _meta
        super().__init_subclass_with_meta__(**options)



"""Mutation tools. """

from .base import Mutation
from .core import (ModelBulkCreationMutationContext,
                   ModelBulkUpdateMutationContext, ModelMutationContext,
                   MutationContext, NodeDeleteMutationContext,
                   NodeMutationContext, NodeUpdateMutationContext)
from .model import (ModelBulkCreationMutation, ModelBulkMutation,
                    ModelBulkUpdateMutation, ModelCreationMutation, ModelMutation,
                    ModelUpdateMutation, get_all_fields)
from .node import NodeDeleteMutation, NodeMutation, NodeUpdateMutation



"""Convert django field to graphene field.  """

from typing import Optional

import django
import graphene
from graphene_django import converter as _converter
from graphene_django.registry import Registry

from ..texttools import camel_case


def construct_enum_from_db_field(field, registry: Optional[Registry] = None):
    """Construct enum field from django db field.

    Args:
        field (django.db.models.Field): Field
        registry (Optional[Registry], optional): Defaults to None. Field registry

    Returns:
        graphene.Enum: Enum field.
    """

    if registry:
        converted = registry.get_converted_field(field)
        if converted:
            return converted
    meta: django.db.models.options.Options = field.model._meta
    name = camel_case("{}_{}".format(meta.object_name, field.name))
    choices = list(_converter.get_choices(field.choices))
    named_choices = [(c[0], c[1]) for c in choices]
    named_choices_descriptions = {c[0]: c[2] for c in choices}

    class EnumWithDescriptionsType(object):
        # pylint:disable=all
        @property
        def description(self):
            return named_choices_descriptions[self.name]

    enum = graphene.Enum(name, list(named_choices),
                         type=EnumWithDescriptionsType)
    converted = enum(description=field.help_text, required=not field.null)
    if registry:
        registry.register_converted_field(field, converted)



"""Graphql middleware for data loader.  """

import logging

from promise import Promise
from promise.dataloader import DataLoader

import graphene_django_tools as gdtools

LOGGER = logging.getLogger(__name__)


def get_model_batch_load_fn(model):
    """Create batch load function for model.  """

    def batch_load_fn(keys):
        LOGGER.debug('load: %s: %s', model, keys)
        result = model.objects.in_bulk(keys)
        return Promise.resolve([result[i] for i in keys])

    return batch_load_fn


def get_model_data_loader(model):
    """Create dataloader for model.  """

    return DataLoader(get_model_batch_load_fn(model))


class DataLoaderMiddleware:
    """Middleware to add `context.get_dataloader` method.  """

    cache_attrname = 'data_loader_cache'

    def resolve(self, resolve_next, parent, info: gdtools.ResolveInfo, **kwargs):
        """Graphene middleware resolve method.  """

        def get_data_loader(model):
            attrname = self.cache_attrname
            if not hasattr(info.context, attrname):
                setattr(info.context, attrname, {})
            cache = getattr(info.context, attrname)
            if model not in cache:
                cache[model] = get_model_data_loader(model)
            return cache[model]

        info.context.get_data_loader = get_data_loader
        return resolve_next(parent, info, **kwargs)



"""Another implement for `graphene.relay.mutation.ClientIDMutation`  """

import re
from typing import Type, Union

import graphene
from graphql import GraphQLError

from . import core
from ..core import get_unmounted_type
from ..interfaces import ClientMutationID
from .base import Mutation


def try_get_node(info: core.ResolveInfo, global_id: str, only_type: Type = False) \
        ->Union[graphene.Node, str]:
    """Fallback to global_id when `graphene.Node.get_node_from_global_id`
        failed.

    Returns:
        Union[graphene.Node, str]: Node or input `global_id`
    """

    try:
        ret = graphene.Node.get_node_from_global_id(
            info=info, global_id=global_id, only_type=only_type)
    except AssertionError:
        ret = None
    if ret is None:
        ret = global_id
    return ret


class NodeMutation(Mutation):
    """Mutation that supports relay.  """
    # pylint: disable=abstract-method

    class Meta:
        abstract = True

    @classmethod
    def __init_subclass_with_meta__(cls, **options):
        options.setdefault('_meta', core.NodeMutationOptions(cls))
        options.setdefault('node_fieldname', 'node')
        options.setdefault('interfaces', ())
        options['interfaces'] += (ClientMutationID,)
        super().__init_subclass_with_meta__(**options)

    @classmethod
    def _construct_meta(cls, **options) -> core.ModelMutationOptions:
        ret = super()._construct_meta(**options)  # type: core.ModelMutationOptions
        ret.node_fieldname = options['node_fieldname']
        return ret

    @classmethod
    def _make_arguments_fields(cls, **options) -> dict:
        fields = super()._make_arguments_fields(**options)
        fields.update(
            __doc__=f'Input data for mutation: {cls.__name__}.',
            client_mutation_id=graphene.ID(
                description='Mutation id for relay, will be returned as-is.'))

        field_objecttype = type(
            re.sub("Input$|$", "Input", cls.__name__),
            (graphene.InputObjectType,),
            fields)
        field = field_objecttype(required=True,
                                 description='Input data this mutation.')
        return dict(input=field)

    @classmethod
    def _make_context(cls, root, info: core.ResolveInfo, **kwargs):
        arguments = kwargs['input']
        return core.NodeMutationContext(root=root,
                                        info=info,
                                        node=None,
                                        options=cls._meta,
                                        arguments=arguments)

    @classmethod
    def premutate(cls, context: core.ModelMutationContext):
        super().premutate(context)
        for k, v in context.arguments.items():
            field = getattr(
                context.options.arguments['input'], '_meta').fields[k]
            context.arguments[k] = \
                cls._convert_id_field_to_node(context, field, v)

    @classmethod
    def _convert_id_field_to_node(cls,
                                  context: core.NodeMutationContext,
                                  field,
                                  value):

        unmounted = get_unmounted_type(field)

        # XXX: maybe replace this with single dispatch.
        if isinstance(unmounted, graphene.types.unmountedtype.UnmountedType):
            method = cls._id2node_unmmounted_instance
        elif issubclass(unmounted, graphene.InputObjectType):
            method = cls._id2node_inputobjecttype
        elif issubclass(unmounted, graphene.Scalar):
            method = cls._id2node_scalar
        else:
            return value

        return method(context, unmounted, value)

    @classmethod
    def _id2node_unmmounted_instance(cls,
                                     context,
                                     unmounted: graphene.types.unmountedtype.UnmountedType,
                                     value):
        ret = value
        if isinstance(unmounted, graphene.List):
            unmounted = unmounted.of_type
            assert isinstance(value, list), type(value)
            ret = [cls._convert_id_field_to_node(
                context, unmounted, i) for i in value]
        return ret

    @classmethod
    def _id2node_inputobjecttype(cls,
                                 context,
                                 unmounted: Type[graphene.Scalar],
                                 value):
        assert isinstance(value, dict), type(value)
        _input_fields = getattr(unmounted, '_meta').fields
        ret = {}
        for k, v in _input_fields.items():
            if k not in value:
                continue
            ret[k] = cls._convert_id_field_to_node(context, v, value[k])
        return ret

    @classmethod
    def _id2node_scalar(cls,
                        context,
                        unmounted: Type[graphene.InputObjectType],
                        value):
        ret = value
        if issubclass(unmounted, graphene.ID):
            ret = try_get_node(context.info, global_id=value)
        return ret

    @classmethod
    def postmutate(cls,
                   context: core.MutationContext,
                   payload: graphene.ObjectType) -> graphene.ObjectType:
        ret = super().postmutate(context, payload)
        ret.client_mutation_id = context.arguments.get("client_mutation_id")
        return ret


class NodeUpdateMutation(NodeMutation):
    """Mutate a node.  """
    # pylint: disable=abstract-method

    class Meta:
        abstract = True

    @classmethod
    def _construct_meta(cls, **options) -> core.NodeUpdateMutationOptions:
        options.setdefault('_meta', core.NodeUpdateMutationOptions(cls))
        options.setdefault('id_fieldname', 'id')

        ret = super()._construct_meta(**options) \
            # type: core.ModelUpdateMutationOptions
        ret.id_fieldname = options['id_fieldname']
        return ret

    @classmethod
    def _make_arguments_fields(cls, **options) -> dict:
        options.setdefault('arguments', {})

        options['arguments'].update(cls._make_node_arguments_fields(**options))
        return super()._make_arguments_fields(**options)

    @classmethod
    def _make_node_arguments_fields(cls, **options) -> dict:
        id_type = graphene.ID(
            required=True,
            description='Node ID for this mutation.')
        return {options['id_fieldname']: id_type}

    @classmethod
    def premutate(cls, context: core.NodeUpdateMutationOptions):
        super().premutate(context)
        node = context.arguments[context.options.id_fieldname]
        if isinstance(node, str) or not node:
            raise GraphQLError(f'No such node: {node}')
        context.node = node


class NodeDeleteMutation(NodeUpdateMutation):
    """Delete a node.  """

    class Meta:
        abstract = True

    @classmethod
    def _construct_meta(cls, **options) -> core.NodeDeleteMutationOptions:
        options.setdefault('_meta', core.NodeDeleteMutationOptions(cls))
        if 'allowed_cls' not in options:
            raise KeyError(
                f'{cls.__name__}: Meta options `allowed_cls` can not be omitted.')

        ret: core.NodeDeleteMutationOptions = super()._construct_meta(**options)
        ret.allowed_cls = options['allowed_cls']
        return ret

    @classmethod
    def mutate(cls, context: core.NodeDeleteMutationContext):
        if not isinstance(context.node, context.options.allowed_cls):
            raise GraphQLError(
                f'Delete not allowed: {type(context.node).__name__}')
        context.node.delete()
        return cls()



"""Tools for use [graphene-django](https://github.com/graphql-python/graphene-django).  """

__version__ = '0.17.0'
__author__ = 'NateScarlet@Gmail.com'

# TODO: Add tests
# TODO: Support other python version
# TODO: Sort field on nodemodel
# TODO: Find documentation for MutationContext.root



"""Convert django field to graphene field for arguments.  """

from functools import singledispatch
from typing import Optional

import django
import graphene
from django import forms
from django.db import models
from graphene.types.json import JSONString
from graphene.types.unmountedtype import UnmountedType
from graphene_django.compat import (ArrayField, HStoreField, JSONField,
                                    RangeField)
from graphene_django.forms.converter import convert_form_field
from graphene_django.registry import Registry

from .enum import construct_enum_from_db_field


def convert_db_field_to_argument(
        field: django.db.models.Field, registry: Optional[Registry] = None) -> UnmountedType:
    """Convert a django db field to graphene field for argument.

    Args:
        field (django.db.models.Field): Field
        registry (Optional[Registry], optional): Defaults to None. Field registry.

    Returns:
        graphene.types.unmounted.UnmmountedType: Graphene field.
    """

    if getattr(field, "choices", None):
        ret = construct_enum_from_db_field(field, registry)
    else:
        ret = _convert_db_field_to_argument(field)
    return ret


@singledispatch
def _convert_db_field_to_argument(field):
    form_field = field.formfield()
    if not form_field:
        return None
    return convert_form_field(form_field)


@_convert_db_field_to_argument.register(models.BooleanField)
def _(field: models.BooleanField):
    return graphene.Boolean(description=field.help_text, required=not field.blank)


@_convert_db_field_to_argument.register(models.ForeignKey)
@_convert_db_field_to_argument.register(models.OneToOneField)
@_convert_db_field_to_argument.register(models.OneToOneRel)
def _(field):
    return graphene.ID(description=getattr(field, 'help_text', ''),
                       required=not getattr(field, 'blank', True))


@_convert_db_field_to_argument.register(models.ManyToManyField)
@_convert_db_field_to_argument.register(models.ManyToManyRel)
@_convert_db_field_to_argument.register(models.ManyToOneRel)
def _(field):
    return graphene.List(graphene.NonNull(graphene.ID), description=getattr(field, 'help_text', ''))


@_convert_db_field_to_argument.register(ArrayField)
@_convert_db_field_to_argument.register(RangeField)
def _(field):
    base_type = _convert_db_field_to_argument(field.base_field)
    if not isinstance(base_type, graphene.NonNull):
        base_type = graphene.NonNull(base_type)
    return graphene.List(base_type, description=field.help_text, required=not field.blank)


@_convert_db_field_to_argument.register(HStoreField)
@_convert_db_field_to_argument.register(JSONField)
def _(field):
    return JSONString(description=field.help_text, required=not field.blank)


@_convert_db_field_to_argument.register(forms.BooleanField)
def _(field):
    # Default `convert_form_field` always return required field.
    # https://github.com/graphql-python/graphene-django/issues/532
    return graphene.Boolean(
        description=field.help_text,
        required=field.required)



"""Python setup script.  """
import os
import re

from setuptools import find_packages, setup

NAME = 'graphene_django_tools'
REQUIRES = [
    'graphene-django~=2.2.0',
    'isodate~=0.6.0'
]

__about__ = {}
with open(os.path.join(os.path.dirname(__file__),
                       NAME, '__about__.py')) as f:
    exec(f.read(), __about__)  # pylint: disable=exec-used


def _get_long_description():
    with open(os.path.join(os.path.dirname(__file__),
                           "README.md"), "r") as f:
        return re.sub(r'(\[.*\]\()\./',
                      lambda matches:
                      (matches[1] +
                       'https://github.com/NateScarlet/graphene-django-tools'
                       '/blob/{}/'.format(__about__['__version__'])),
                      f.read())


setup(
    name=NAME,
    version=__about__['__version__'],
    author=__about__['__author__'],
    author_email=__about__['__author__'],
    url='https://github.com/NateScarlet/graphene-django-tools',
    packages=find_packages(),
    package_data={},
    install_requires=REQUIRES,
    description='Tools for use [graphene-django](https://github.com/graphql-python/graphene-django)',
    long_description=_get_long_description(),
    long_description_content_type="text/markdown",
    classifiers=['Framework :: Django :: 2.1',
                 'Framework :: Django :: 2.2',
                 'Intended Audience :: Developers',
                 'License :: OSI Approved :: MIT License',
                 'Programming Language :: Python :: 3.7',
                 'Topic :: Software Development :: Libraries :: Python Modules',
                 ],
)




"""Another implementation for `graphene.Mutation`  """

import re
from collections import OrderedDict
from typing import Any, Dict, Type

import graphene
import graphene_django
import graphene_django.forms.mutation
from graphene.types.utils import yank_fields_from_attrs
from graphene.utils.get_unbound_function import get_unbound_function
from graphene.utils.props import props
from graphql.execution.base import ResolveInfo

from . import core
from ..core import handle_resolve_error


class Mutation(graphene.ObjectType):
    """Base mutation class.  """

    class Meta:
        abstract = True

    @classmethod
    def __init_subclass_with_meta__(cls, **options):
        # pylint: disable=W0221
        options['_meta'] = cls._construct_meta(**options)
        options.setdefault('name', cls.__name__)
        options['name'] = re.sub("Payload$|$", "Payload", options['name'])

        allowed_keys = ('_meta', 'name', 'description',
                        'interfaces', 'possible_types', 'default_resolver')
        options = {k: v for k, v in options.items() if k in allowed_keys}
        super().__init_subclass_with_meta__(**options)

    @classmethod
    def _construct_meta(cls, **options) -> graphene_django.forms.mutation.MutationOptions:
        ret = options.get(
            '_meta', graphene_django.forms.mutation.MutationOptions(cls))
        ret.output = cls._make_payload_objecttype(**options)
        ret.resolver = cls._make_resolver(**options)
        ret.arguments = cls._make_arguments_fields(**options)
        ret.fields = cls._make_payload_fields(**options)
        return ret

    @classmethod
    def _make_payload_objecttype(cls, **options) -> Type[graphene.ObjectType]:
        ret = options.get('output', getattr(cls, 'Output', cls))
        assert issubclass(ret, graphene.ObjectType), (cls, type(ret), options)
        return ret

    @classmethod
    def _make_payload_fields(cls, **options) -> OrderedDict:
        if 'output' in options:
            return options['output'].fields

        ret = getattr(options.get('_meta'), 'fields', None) or OrderedDict()
        for base in reversed(cls.__mro__):
            _fields = yank_fields_from_attrs(base.__dict__,
                                             _as=graphene.Field)
            ret.update(_fields)
        return ret

    @classmethod
    def _make_arguments_fields(cls, **options) -> OrderedDict:
        options.setdefault('interfaces', ())
        options.setdefault('arguments', {})

        def _get_arguments(interface):
            if hasattr(interface, 'Arguments'):
                return props(getattr(interface, 'Arguments'))
            return {}

        ret = options['arguments']
        for i in options['interfaces'] + (cls,):
            ret.update(_get_arguments(i))
        return ret

    @classmethod
    def _make_resolver(cls, **options):
        # pylint: disable=unused-argument
        return get_unbound_function(cls._resolver)

    @classmethod
    def _make_context(cls, root, info: ResolveInfo, **kwargs):
        return core.MutationContext(root=root, info=info, options=cls._meta, arguments=kwargs)

    @classmethod
    def _resolver(cls, *args, **kwargs: Dict[str, Any]):
        """Resolve the graphQL mutate.

        Only override this on abstract object.
        For schema, use `preform_mutate` instead.
        """

        try:
            context = cls._make_context(*args, **kwargs)
            cls.premutate(context)
            payload = cls.mutate(context)
            payload = cls.postmutate(context, payload)
            return payload
        except:
            handle_resolve_error()
            raise

    @classmethod
    def mutate(cls, context: core.MutationContext) \
            -> graphene.ObjectType:
        """Do the mutation.  """

        raise NotImplementedError(
            f'Method not implemented: {cls.__name__}.mutate')

    @classmethod
    def premutate(cls, context: core.MutationContext):
        """Actions before mutation perform.

        raise a error to abort the mutation.
        """

    @classmethod
    def postmutate(cls, context: core.MutationContext,
                   payload: graphene.ObjectType) -> graphene.ObjectType:
        """Actions after mutation performed.

        Returns:
            graphene.ObjectType: Result value.
        """
        # pylint:disable=unused-argument
        assert isinstance(
            payload, graphene.ObjectType), \
            f'Wrong payload type: {type(payload)}'
        return payload

    @classmethod
    def Field(cls, **kwargs) -> graphene.Field:
        """Create graphene Field for the mutation.  """
        # pylint: disable=invalid-name
        assert issubclass(cls._meta.output, graphene.ObjectType), type(
            cls._meta.output)
        return graphene.Field(
            type=cls._meta.output,
            args=cls._meta.arguments,
            resolver=cls._meta.resolver,
            name=kwargs.get('name'),
            required=kwargs.get('required', False),
            default_value=kwargs.get('default_value'),
            description=kwargs.get('description', cls.__doc__),
            deprecation_reason=kwargs.get('deprecation_reason'),)



"""Example schema for django auth.  """

import graphene
from django.contrib.auth import authenticate, get_user_model, login, logout
from django.contrib.auth.password_validation import validate_password

import graphene_django_tools as gdtools

User = get_user_model()  # pylint: disable=invalid-name


class UserMutation(gdtools.ModelMutation):
    """Additional actions for user.  """

    class Meta:
        abstract = True

    @classmethod
    def premutate(cls, context: gdtools.ModelMutationContext):

        super().premutate(context)
        password = context.mapping.get('password')
        if password:
            validate_password(password)

    @classmethod
    def postmutate(cls,
                   context: gdtools.ModelMutationContext,
                   payload: graphene.ObjectType) -> graphene.ObjectType:

        password = context.mapping.get('password')
        if password:
            context.instance.set_password(password)
        context.instance.full_clean()
        return super().postmutate(context, payload)


class CreateUser(UserMutation, gdtools.ModelCreationMutation):
    """Create user.  """

    class Meta:
        model = User
        fields = ('username', 'password', 'first_name', 'last_name', 'email')
        require = ('username', 'password')


class UpdateUser(UserMutation, gdtools.ModelUpdateMutation):
    """Update user.  """

    class Meta:
        model = User
        fields = ('username', 'password', 'first_name', 'last_name', 'email')


class Login(gdtools.NodeMutation):
    """Login current user.  """

    class Arguments:
        username = graphene.String(required=True)
        password = graphene.String(required=True)

    user = gdtools.ModelField(User)

    @classmethod
    def mutate(cls, context: gdtools.ModelMutationContext):
        request = context.info.context
        user = authenticate(**context.arguments)
        if not user:
            raise ValueError('Login failed.')
        login(request, user)
        return cls(user=user)


class Logout(gdtools.NodeMutation):
    """Logout current user.  """

    @classmethod
    def mutate(cls, context: gdtools.ModelMutationContext):
        logout(context.info.context)
        return cls()



"""Mutation tools. """

from collections import OrderedDict
from typing import Any, List

import graphene
from django import db
from graphene.types.mutation import MutationOptions
from graphql.execution.base import ResolveInfo

from dataclasses import dataclass

# pylint: disable=too-few-public-methods


class NodeMutationOptions(MutationOptions):
    """`Meta` for `NodeMutation`.  """

    node_fieldname = None  # type: str


class NodeUpdateMutationOptions(NodeMutationOptions):
    """`Meta` for `NodeUpdateMutation`.  """

    id_fieldname = None  # type: str


class NodeDeleteMutationOptions(NodeUpdateMutationOptions):
    """`Meta` for `NodeDeleteMutation`.  """

    allowed_cls: tuple = ()


class ModelMutationOptions(NodeMutationOptions):
    """`Meta` for `ModelMutation`.  """

    model = None  # type: django.db.models.Model
    fields = ()  # type: Union[Callable, Iterable]
    require = ()  # type: tuple[str]
    exclude = ()  # type: tuple[str]
    require_mapping = True


class ModelUpdateMutationOptions(NodeUpdateMutationOptions, ModelMutationOptions):
    """`Meta` for `ModelUpdateMutation`.  """


@dataclass
class MutationContext:
    """Tuple data for mutation context.  """

    root: Any  # XXX: Not found documentation for this.
    info: ResolveInfo
    options: MutationOptions
    arguments: dict


@dataclass
class NodeMutationContext(MutationContext):

    node: graphene.relay.Node
    options: NodeMutationOptions


@dataclass
class NodeUpdateMutationContext(MutationContext):
    options: NodeUpdateMutationOptions


@dataclass
class NodeDeleteMutationContext(MutationContext):
    options: NodeDeleteMutationOptions


@dataclass
class ModelMutationContext(NodeMutationContext):
    """Tuple data for model mutation context.  """

    mapping: dict
    instance: graphene.ObjectType = None
    options: ModelMutationOptions


@dataclass
class ModelBulkCreationMutationContext(MutationContext):
    """Tuple data for model bulk creation mutation context.  """

    data: List[dict]


@dataclass
class ModelBulkUpdateMutationContext(MutationContext):
    """Tuple data for model bulk update mutation context.  """
    mapping: dict
    query_set: db.models.QuerySet = None


def _sorted_dict(obj: dict, key=None, reverse=False)->OrderedDict:
    return OrderedDict(sorted(obj.items(), key=key, reverse=reverse))



"""Django graphene model mutation"""

import re


def camel_case(text: str, is_lower=False) -> str:
    """Convert text to camel case.

    Args:
        text (str)

    Returns:
        str
    """

    components = text.split('_')
    first = components[0].lower() if is_lower else components[0].title()
    return first + ''.join(x.title() for x in components[1:])


def snake_case(text: str) -> str:
    """Convert text to snake case.

    Args:
        text (str)

    Returns:
        str
    """

    ret = text
    ret = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', ret)
    ret = re.sub('([a-z0-9])([A-Z])', r'\1_\2', ret)
    ret = ret.lower()
    return ret
